# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Entry points and functionality for importing models.

This module contains entry points for the psamm-import programs and
functionality to assist in importing external file formats, converting
them into proper native models and writing those models to files.
"""

from __future__ import print_function, unicode_literals

import sys
import os
import argparse
import logging
import math
import re
import json
import codecs
from itertools import product
from collections import OrderedDict, Counter
from decimal import Decimal

import yaml
import pkg_resources
from six import iteritems, itervalues, text_type
from six.moves.urllib.request import urlopen
from six.moves.urllib.parse import quote as url_quote

from .datasource.native import ModelWriter
from .datasource.reaction import (parse_reaction,
                                  ParseError as ReactionParseError)
from .datasource.entry import DictCompartmentEntry
from .datasource import sbml
from .expression import boolean
from . import formula

from .util import mkdir_p

# Threshold for putting reactions into subsystem files
_MAX_REACTION_COUNT = 3

# Threshold for converting reactions into dictionary representation.
_MAX_REACTION_LENGTH = 10

logger = logging.getLogger(__name__)


class ImportError(Exception):
    """Exception used to signal a general import error."""


class ModelLoadError(ImportError):
    """Exception used to signal an error loading the model files."""


class ParseError(ImportError):
    """Exception used to signal an error parsing the model files."""


class Importer(object):
    """Base importer class."""

    def _try_parse_formula(self, compound_id, s):
        """Try to parse the given compound formula string.

        Logs a warning if the formula could not be parsed.
        """
        s = s.strip()
        if s == '':
            return None

        try:
            # Do not return the parsed formula. For now it is better to keep
            # the original formula string unchanged in all cases.
            formula.Formula.parse(s)
        except formula.ParseError:
            logger.warning('Unable to parse compound formula {}: {}'.format(
                compound_id, s))

        return s

    def _try_parse_reaction(self, reaction_id, s,
                            parser=parse_reaction, **kwargs):
        """Try to parse the given reaction equation string.

        Returns the parsed Reaction object, or raises an error if the reaction
        could not be parsed.
        """
        try:
            return parser(s, **kwargs)
        except ReactionParseError as e:
            if e.indicator is not None:
                logger.error('{}\n{}\n{}'.format(
                    str(e), s, e.indicator))
            raise ParseError('Unable to parse reaction {}: {}'.format(
                reaction_id, s))

    def _try_parse_gene_association(self, reaction_id, s):
        """Try to parse the given gene association rule.

        Logs a warning if the association rule could not be parsed and returns
        the original string. Otherwise, returns the boolean.Expression object.
        """
        s = s.strip()
        if s == '':
            return None

        try:
            return boolean.Expression(s)
        except boolean.ParseError as e:
            msg = 'Failed to parse gene association for {}: {}'.format(
                reaction_id, text_type(e))
            if e.indicator is not None:
                msg += '\n{}\n{}'.format(s, e.indicator)
            logger.warning(msg)

        return s


# Define custom dict representers for YAML
# This allows reading/writing Python OrderedDicts in the correct order.
# See: https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts  # noqa
def _dict_representer(dumper, data):
    return dumper.represent_dict(iteritems(data))


def _dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


def _set_representer(dumper, data):
    return dumper.represent_list(iter(data))


def _boolean_expression_representer(dumper, data):
    return dumper.represent_unicode(text_type(data))


def _decimal_representer(dumper, data):
    # Code from float_representer in PyYAML.
    if data % 1 == 0:
        return dumper.represent_int(int(data))
    elif math.isnan(data):
        value = '.nan'
    elif data == float('inf'):
        value = '.inf'
    elif data == float('-inf'):
        value = '-.inf'
    else:
        value = text_type(data).lower()
        if '.' not in value and 'e' in value:
            value = value.replace('e', '.0e', 1)
    return dumper.represent_scalar('tag:yaml.org,2002:float', value)


def get_default_compartment(model):
    """Return what the default compartment should be set to.

    If some compounds have no compartment, unique compartment
    name is returned to avoid collisions.
    """
    default_compartment = 'c'
    default_key = set()
    for reaction in model.reactions:
        equation = reaction.equation
        if equation is None:
            continue

        for compound, _ in equation.compounds:
            default_key.add(compound.compartment)

    if None in default_key and default_compartment in default_key:
        suffix = 1
        while True:
            new_key = '{}_{}'.format(default_compartment, suffix)
            if new_key not in default_key:
                default_compartment = new_key
                break
            suffix += 1

    if None in default_key:
        logger.warning(
            'Compound(s) found without compartment, default'
            ' compartment is set to {}.'.format(default_compartment))
    return default_compartment


def detect_best_flux_limit(model):
    """Detect the best default flux limit to use for model output.

    The default flux limit does not change the model but selecting a good
    value reduced the amount of output produced and reduces clutter in the
    output files.
    """
    flux_limit_count = Counter()

    for reaction in model.reactions:
        if reaction.id not in model.limits:
            continue

        equation = reaction.properties['equation']
        if equation is None:
            continue

        _, lower, upper = model.limits[reaction.id]
        if upper is not None and upper > 0 and equation.direction.forward:
            flux_limit_count[upper] += 1
        if lower is not None and -lower > 0 and equation.direction.reverse:
            flux_limit_count[-lower] += 1

    if len(flux_limit_count) == 0:
        return None

    best_flux_limit, _ = flux_limit_count.most_common(1)[0]
    return best_flux_limit


def reactions_to_files(model, dest, writer, split_subsystem):
    """Turn the reaction subsystems into their own files.

    If a subsystem has a number of reactions over the threshold, it gets its
    own YAML file. All other reactions, those that don't have a subsystem or
    are in a subsystem that falls below the threshold, get added to a common
    reaction file.

    Args:
        model: :class:`psamm_import.model.MetabolicModel`.
        dest: output path for model files.
        writer: :class:`psamm.datasource.native.ModelWriter`.
        split_subsystem: Divide reactions into multiple files by subsystem.
    """
    def safe_file_name(origin_name):
        safe_name = re.sub(
            r'\W+', '_', origin_name, flags=re.UNICODE)
        safe_name = re.sub(
            r'_+', '_', safe_name.lower(), flags=re.UNICODE)
        safe_name = safe_name.strip('_')
        return safe_name

    common_reactions = []
    reaction_files = []
    if not split_subsystem:
        common_reactions = sorted(model.reactions, key=lambda r: r.id)
        if len(common_reactions) > 0:
            reaction_file = 'reactions.yaml'
            with open(os.path.join(dest, reaction_file), 'w') as f:
                writer.write_reactions(f, common_reactions)
            reaction_files.append(reaction_file)
    else:
        subsystems = {}
        for reaction in sorted(model.reactions, key=lambda r: r.id):
            if 'subsystem' in reaction.properties:
                subsystem_file = safe_file_name(
                    reaction.properties['subsystem'])
                subsystems.setdefault(subsystem_file, []).append(reaction)
            else:
                common_reactions.append(reaction)

        subsystem_folder = 'reactions'
        sub_existance = False
        for subsystem_file, reactions in iteritems(subsystems):
            if len(reactions) < _MAX_REACTION_COUNT:
                for reaction in reactions:
                    common_reactions.append(reaction)
            else:
                if len(reactions) > 0:
                    mkdir_p(os.path.join(dest, subsystem_folder))
                    subsystem_file = os.path.join(
                        subsystem_folder, '{}.yaml'.format(subsystem_file))

                    with open(os.path.join(dest, subsystem_file), 'w') as f:
                        writer.write_reactions(f, reactions)
                    reaction_files.append(subsystem_file)
                    sub_existance = True

        reaction_files.sort()
        if sub_existance:
            reaction_file = os.path.join(
                subsystem_folder, 'other_reactions.yaml')
        else:
            reaction_file = 'reactions.yaml'
        if len(common_reactions) > 0:
            with open(os.path.join(dest, reaction_file), 'w') as f:
                writer.write_reactions(f, common_reactions)
            reaction_files.append(reaction_file)

    return reaction_files


def _get_output_limit(limit, default_limit):
    """Return limit to output given default limit."""
    return limit if limit != default_limit else None


def _generate_limit_items(lower, upper):
    """Yield key, value pairs for limits dictionary.

    Yield pairs of key, value where key is ``lower``, ``upper`` or ``fixed``.
    A key, value pair is emitted if the bounds are not None.
    """
    # Use value + 0 to convert any -0.0 to 0.0 which looks better.
    if lower is not None and upper is not None and lower == upper:
        yield 'fixed', upper + 0
    else:
        if lower is not None:
            yield 'lower', lower + 0
        if upper is not None:
            yield 'upper', upper + 0


def model_exchange(model):
    """Return exchange definition as YAML dict."""
    # Determine the default flux limits. If the value is already at the
    # default it does not need to be included in the output.
    lower_default, upper_default = None, None
    if model.default_flux_limit is not None:
        lower_default = -model.default_flux_limit
        upper_default = model.default_flux_limit

    compounds = []
    for compound, reaction_id, lower, upper in sorted(
            itervalues(model.exchange)):
        d = OrderedDict([('id', compound.name)])
        if reaction_id is not None:
            d['reaction'] = reaction_id

        lower = _get_output_limit(lower, lower_default)
        upper = _get_output_limit(upper, upper_default)
        d.update(_generate_limit_items(lower, upper))

        compounds.append(d)

    return OrderedDict([('compounds', compounds)])


def model_reaction_limits(model):
    """Yield model reaction limits as YAML dicts."""
    for reaction in sorted(model.reactions, key=lambda r: r.id):
        equation = reaction.properties.get('equation')
        if equation is None:
            continue

        # Determine the default flux limits. If the value is already at the
        # default it does not need to be included in the output.
        lower_default, upper_default = None, None
        if model.default_flux_limit is not None:
            if equation.direction.reverse:
                lower_default = -model.default_flux_limit
            else:
                lower_default = 0.0

            if equation.direction.forward:
                upper_default = model.default_flux_limit
            else:
                upper_default = 0.0

        lower_flux, upper_flux = None, None
        if reaction.id in model.limits:
            _, lower, upper = model.limits[reaction.id]
            lower_flux = _get_output_limit(lower, lower_default)
            upper_flux = _get_output_limit(upper, upper_default)

        if lower_flux is not None or upper_flux is not None:
            d = OrderedDict([('reaction', reaction.id)])
            d.update(_generate_limit_items(lower_flux, upper_flux))

            yield d


def infer_compartment_entries(model):
    """Infer compartment entries for model based on reaction compounds."""
    compartment_ids = set()
    for reaction in model.reactions:
        equation = reaction.equation
        if equation is None:
            continue

        for compound, _ in equation.compounds:
            compartment = compound.compartment
            if compartment is None:
                compartment = model.default_compartment

            if compartment is not None:
                compartment_ids.add(compartment)

    for compartment in compartment_ids:
        if compartment in model.compartments:
            continue

        entry = DictCompartmentEntry(dict(id=compartment))
        model.compartments.add_entry(entry)


def infer_compartment_adjacency(model):
    """Infer compartment adjacency for model based on reactions."""
    def reaction_compartments(seq):
        for compound, _ in seq:
            compartment = compound.compartment
            if compartment is None:
                compartment = model.default_compartment

            if compartment is not None:
                yield compartment

    for reaction in model.reactions:
        equation = reaction.equation
        if equation is None:
            continue

        left = reaction_compartments(equation.left)
        right = reaction_compartments(equation.right)
        for c1, c2 in product(left, right):
            if c1 == c2:
                continue
            model.compartment_boundaries.add((c1, c2))
            model.compartment_boundaries.add((c2, c1))


def count_genes(model):
    """Count the number of distinct genes in model reactions."""
    genes = set()
    for reaction in model.reactions:
        if reaction.genes is None:
            continue

        if isinstance(reaction.genes, boolean.Expression):
            genes.update(v.symbol for v in reaction.genes.variables)
        else:
            gene_exp = boolean.Expression(reaction.genes)
            genes.update(v.symbol for v in gene_exp.variables)
            # genes.update(reaction.genes)

    return len(genes)


def write_yaml_model(model, dest='.', convert_exchange=True,
                     split_subsystem=True):
    """Write the given NativeModel to YAML files in dest folder.

    The parameter ``convert_exchange`` indicates whether the exchange reactions
    should be converted automatically to an exchange file.
    """
    yaml.SafeDumper.add_representer(OrderedDict, _dict_representer)
    yaml.SafeDumper.add_representer(set, _set_representer)
    yaml.SafeDumper.add_representer(frozenset, _set_representer)
    yaml.SafeDumper.add_representer(
        boolean.Expression, _boolean_expression_representer)
    yaml.SafeDumper.add_representer(Decimal, _decimal_representer)

    yaml.SafeDumper.ignore_aliases = lambda *args: True

    yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                         _dict_constructor)

    yaml_args = {'default_flow_style': False,
                 'encoding': 'utf-8',
                 'allow_unicode': True,
                 'width': 79}

    # The ModelWriter from PSAMM is not yet able to write the full model but
    # only reactions and compounds.
    writer = ModelWriter()

    with open(os.path.join(dest, 'compounds.yaml'), 'w+') as f:
        writer.write_compounds(f, sorted(model.compounds, key=lambda c: c.id))

    if model.default_flux_limit is None:
        model.default_flux_limit = detect_best_flux_limit(model)

    if model.extracellular_compartment is None:
        model.extracellular_compartment = (
            sbml.detect_extracellular_compartment(model))

    if model.default_compartment is None:
        model.default_compartment = get_default_compartment(model)

    if model.default_flux_limit is not None:
        logger.info('Using default flux limit of {}'.format(
            model.default_flux_limit))

    if convert_exchange:
        logger.info('Converting exchange reactions to exchange file')
        sbml.convert_exchange_to_compounds(model)

    if len(model.compartments) == 0:
        infer_compartment_entries(model)
        logger.info('Inferred {} compartments: {}'.format(
            len(model.compartments),
            ', '.join(c.id for c in model.compartments)))

    if len(model.compartments) != 0 and len(model.compartment_boundaries) == 0:
        infer_compartment_adjacency(model)

    reaction_files = reactions_to_files(model, dest, writer, split_subsystem)

    if len(model.exchange) > 0:
        with open(os.path.join(dest, 'exchange.yaml'), 'w+') as f:
            yaml.safe_dump(model_exchange(model), f, **yaml_args)

    reaction_limits = list(model_reaction_limits(model))
    if len(reaction_limits) > 0:
        with open(os.path.join(dest, 'limits.yaml'), 'w+') as f:
            yaml.safe_dump(reaction_limits, f, **yaml_args)

    model_d = OrderedDict()
    if model.name is not None:
        model_d['name'] = model.name

    if model.biomass_reaction is not None:
        model_d['biomass'] = model.biomass_reaction
    if model.default_flux_limit is not None:
        model_d['default_flux_limit'] = model.default_flux_limit
    if model.extracellular_compartment != 'e':
        model_d['extracellular'] = model.extracellular_compartment
    if model.default_compartment != 'c':
        model_d['default_compartment'] = model.default_compartment

    if len(model.compartments) > 0:
        adjacency = {}
        for c1, c2 in model.compartment_boundaries:
            adjacency.setdefault(c1, set()).add(c2)
            adjacency.setdefault(c2, set()).add(c1)

        compartment_list = []
        for compartment in sorted(model.compartments, key=lambda c: c.id):
            adjacent = adjacency.get(compartment.id)
            if adjacent is not None and len(adjacent) == 1:
                adjacent = next(iter(adjacent))
            compartment_list.append(writer.convert_compartment_entry(
                compartment, adjacent))

        model_d['compartments'] = compartment_list

    model_d['compounds'] = [{'include': 'compounds.yaml'}]
    model_d['reactions'] = []
    for reaction_file in reaction_files:
        model_d['reactions'].append({'include': reaction_file})

    if len(model.exchange) > 0:
        model_d['exchange'] = [{'include': 'exchange.yaml'}]

    if len(reaction_limits) > 0:
        model_d['limits'] = [{'include': 'limits.yaml'}]

    with open(os.path.join(dest, 'model.yaml'), 'w+') as f:
        yaml.safe_dump(model_d, f, **yaml_args)


def main(importer_class=None, args=None):
    """Entry point for import program.

    If the ``args`` are provided, these should be a list of strings that will
    be used instead of ``sys.argv[1:]``. This is mostly useful for testing.
    """
    parser = argparse.ArgumentParser(
        description='Import from external model formats')
    parser.add_argument('--source', metavar='path', default='.',
                        help='Source directory or file')
    parser.add_argument('--dest', metavar='path', default='.',
                        help='Destination directory (default is ".")')
    parser.add_argument('--no-exchange', action='store_true',
                        help=('Disable importing exchange reactions as'
                              ' exchange compound file.'))
    parser.add_argument('--split-subsystem', action='store_true',
                        help='Enable splitting reaction files by subsystem')
    parser.add_argument('--merge-compounds', action='store_true',
                        help=('Merge identical compounds occuring in various'
                              ' compartments.'))
    parser.add_argument('--force', action='store_true',
                        help='Enable overwriting model files')
    parser.add_argument('--model-name', type=str,
                        help=('specify model name if multiple models '
                              'are stored in the .mat file'))

    if importer_class is None:
        parser.add_argument(
            'format', help='Format to import ("list" to see all)')

    args = parser.parse_args(args)

    # Set up logging for the command line interface
    if 'PSAMM_DEBUG' in os.environ:
        level = getattr(logging, os.environ['PSAMM_DEBUG'].upper(), None)
        if level is not None:
            logging.basicConfig(level=level)
    else:
        logging.basicConfig(
            level=logging.INFO, format='%(levelname)s: %(message)s')

    if importer_class is None:
        # Discover all available model importers
        importers = {}
        for importer_entry in pkg_resources.iter_entry_points(
                'psamm.importer'):
            canonical = importer_entry.name.lower()
            if canonical not in importers:
                importers[canonical] = importer_entry
            else:
                logger.warning('Importer {} was found more than once!'.format(
                    importer_entry.name))

        # Print list of importers
        if args.format in ('list', 'help'):
            if len(importers) == 0:
                logger.error('No importers found!')
            else:
                importer_classes = []
                for name, entry in iteritems(importers):
                    importer_class = entry.load()
                    title = getattr(importer_class, 'title', None)
                    generic = getattr(importer_class, 'generic', False)
                    if title is not None:
                        importer_classes.append(
                            (title, generic, name, importer_class))

                print('Generic importers:')
                for title, _, name, importer_class in sorted(
                        c for c in importer_classes if c[1]):
                    print('{:<12}  {}'.format(name, title))

                print()
                print('Model-specific importers:')
                for title, _, name, importer_class in sorted(
                        c for c in importer_classes if not c[1]):
                    print('{:<12}  {}'.format(name, title))
            sys.exit(0)

        importer_name = args.format.lower()
        if importer_name not in importers:
            logger.error('Importer {} not found!'.format(importer_name))
            logger.info('Use "list" to see available importers.')
            sys.exit(-1)

        importer_class = importers[importer_name].load()

    importer = importer_class()

    try:
        if importer.name == 'matlab' and args.model_name is not None:
            model = importer.import_model(args.source, args.model_name)
        else:
            model = importer.import_model(args.source)
    except ModelLoadError as e:
        logger.error('Failed to load model!', exc_info=True)
        importer.help()
        parser.error(text_type(e))
    except ParseError as e:
        logger.error('Failed to parse model!', exc_info=True)
        logger.error(text_type(e))
        sys.exit(-1)
    except Exception:
        importer.help()

    if args.merge_compounds:
        compounds_before = len(model.compounds)
        sbml.merge_equivalent_compounds(model)
        if len(model.compounds) < compounds_before:
            logger.info(
                'Merged {} compound entries into {} entries by'
                ' removing duplicates in various compartments'.format(
                    compounds_before, len(model.compounds)))

    print('Model: {}'.format(model.name))
    print('- Biomass reaction: {}'.format(model.biomass_reaction))
    print('- Compartments: {}'.format(len(model.compartments)))
    print('- Compounds: {}'.format(len(model.compounds)))
    print('- Reactions: {}'.format(len(model.reactions)))
    print('- Genes: {}'.format(count_genes(model)))

    # Check if dest directory is empty. If we get an error assume that the
    # directory does not exist.
    dest_is_empty = False
    try:
        dest_is_empty = len(os.listdir(args.dest)) == 0
    except OSError:
        dest_is_empty = True

    if not dest_is_empty:
        if not args.force:
            logger.error('Destination directory is not empty. Use --force'
                         ' option to proceed anyway, overwriting any existing'
                         ' files in {}'.format(args.dest))
            return 1
        else:
            logger.warning('Destination directory is not empty, overwriting'
                           ' existing files in {}'.format(args.dest))

    # Create destination directory if not exists
    dest = args.dest
    mkdir_p(dest)

    convert_exchange = not args.no_exchange
    write_yaml_model(model, dest, convert_exchange=convert_exchange,
                     split_subsystem=args.split_subsystem)


def main_bigg(args=None, urlopen=urlopen):
    """Entry point for BiGG import program.

    If the ``args`` are provided, these should be a list of strings that will
    be used instead of ``sys.argv[1:]``. This is mostly useful for testing.
    """
    parser = argparse.ArgumentParser(
        description='Import from BiGG database')
    parser.add_argument('--dest', metavar='path', default='.',
                        help='Destination directory (default is ".")')
    parser.add_argument('--no-exchange', action='store_true',
                        help=('Disable importing exchange reactions as'
                              ' exchange compound file.'))
    parser.add_argument('--split-subsystem', action='store_true',
                        help='Enable splitting reaction files by subsystem')
    parser.add_argument('--merge-compounds', action='store_true',
                        help=('Merge identical compounds occuring in various'
                              ' compartments.'))
    parser.add_argument('--force', action='store_true',
                        help='Enable overwriting model files')
    parser.add_argument('id', help='BiGG model to import ("list" to see all)')

    args = parser.parse_args(args)

    # Set up logging for the command line interface
    if 'PSAMM_DEBUG' in os.environ:
        level = getattr(logging, os.environ['PSAMM_DEBUG'].upper(), None)
        if level is not None:
            logging.basicConfig(level=level)
    else:
        logging.basicConfig(
            level=logging.INFO, format='%(levelname)s: %(message)s')

    # Print list of available models
    if args.id == 'list':
        print('Available models:')
        f = urlopen('http://bigg.ucsd.edu/api/v2/models')
        doc = json.loads(f.read().decode('utf-8'))
        results = doc['results']
        id_width = min(max(len(result['bigg_id']) for result in results), 16)
        for result in sorted(results, key=lambda x: x.get('organism')):
            print('{} {}'.format(
                result.get('bigg_id').ljust(id_width), result.get('organism')))
        return 0

    importer_entry = None
    try:
        importer_entry = next(
            pkg_resources.iter_entry_points('psamm.importer', 'JSON'))
    except StopIteration:
        logger.error('Failed to locate the COBRA JSON model importer!')
        sys.exit(-1)

    importer_class = importer_entry.load()
    importer = importer_class()

    try:
        f = urlopen(
            'http://bigg.ucsd.edu/api/v2/models/{}/download'.format(
                url_quote(args.id)))
        model = importer.import_model(codecs.getreader('utf-8')(f))
    except ModelLoadError as e:
        logger.error('Failed to load model!', exc_info=True)
        importer.help()
        parser.error(text_type(e))
    except ParseError as e:
        logger.error('Failed to parse model!', exc_info=True)
        logger.error(text_type(e))
        sys.exit(-1)

    if args.merge_compounds:
        compounds_before = len(model.compounds)
        sbml.merge_equivalent_compounds(model)
        if len(model.compounds) < compounds_before:
            logger.info(
                'Merged {} compound entries into {} entries by'
                ' removing duplicates in various compartments'.format(
                    compounds_before, len(model.compounds)))

    print('Model: {}'.format(model.name))
    print('- Biomass reaction: {}'.format(model.biomass_reaction))
    print('- Compartments: {}'.format(len(model.compartments)))
    print('- Compounds: {}'.format(len(model.compounds)))
    print('- Reactions: {}'.format(len(model.reactions)))
    print('- Genes: {}'.format(count_genes(model)))

    # Check if dest directory is empty. If we get an error assume that the
    # directory does not exist.
    dest_is_empty = False
    try:
        dest_is_empty = len(os.listdir(args.dest)) == 0
    except OSError:
        dest_is_empty = True

    if not dest_is_empty:
        if not args.force:
            logger.error('Destination directory is not empty. Use --force'
                         ' option to proceed anyway, overwriting any existing'
                         ' files in {}'.format(args.dest))
            return 1
        else:
            logger.warning('Destination directory is not empty, overwriting'
                           ' existing files in {}'.format(args.dest))

    # Create destination directory if not exists
    dest = args.dest
    mkdir_p(dest)

    convert_exchange = not args.no_exchange
    write_yaml_model(model, dest, convert_exchange=convert_exchange,
                     split_subsystem=args.split_subsystem)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import sys
import argparse

from ..datasource import sbml
from ..command import MetabolicMixin, Command


class SBMLExport(MetabolicMixin, Command):
    """Export model as SBML level 3 file."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            'file', type=argparse.FileType('w'), nargs='?', default=sys.stdout,
            help='File path for writing the SBML file'
                 ' (use "-" for standard output)')
        parser.add_argument(
            '--pretty', action='store_true',
            help='Format output for readability')
        super(SBMLExport, cls).init_parser(parser)

    def run(self):
        writer = sbml.SBMLWriter()
        writer.write_model(self._args.file, self._model, self._args.pretty)



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import print_function

import sys
from setuptools import setup, find_packages
import pkg_resources

# Read long description
with open('README.rst') as f:
    long_description = f.read()

# Test whether psamm-import is currently installed. Since the psamm-import
# functionality was moved to this package (except Excel importers), only newer
# versions of psamm-import are compatible with recent versions of PSAMM.
try:
    pkg_resources.get_distribution('psamm-import <= 0.15.2')
except (pkg_resources.DistributionNotFound,
        pkg_resources.VersionConflict):
    pass
else:
    msg = (
        'Please upgrade or uninstall psamm-import before upgrading psamm:\n'
        '$ pip install --upgrade psamm-import\n'
        ' OR\n'
        '$ pip uninstall psamm-import'
        '\n\n'
        ' The functionality of the psamm-import package has been moved into'
        ' the psamm package, and the psamm-import package now only contains'
        ' the model-specific Excel importers.')
    print(msg, file=sys.stderr)
    sys.exit(1)


setup(
    name='psamm',
    version='1.1.2',
    description='PSAMM metabolic modeling tools',
    maintainer='Jon Lund Steffensen',
    maintainer_email='jon_steffensen@uri.edu',
    url='https://github.com/zhanglab/psamm',
    license='GNU GPLv3+',

    long_description=long_description,

    classifiers=[
        'Development Status :: 4 - Beta',
        (
            'License :: OSI Approved :: '
            'GNU General Public License v3 or later (GPLv3+)'),
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    packages=find_packages(),

    entry_points='''
        [console_scripts]
        psamm-model = psamm.command:main
        psamm-sbml-model = psamm.command:main_sbml
        psamm-list-lpsolvers = psamm.lpsolver.generic:list_solvers
        psamm-import = psamm.importer:main
        psamm-import-bigg = psamm.importer:main_bigg

        [psamm.commands]
        chargecheck = psamm.commands.chargecheck:ChargeBalanceCommand
        console = psamm.commands.console:ConsoleCommand
        dupcheck = psamm.commands.duplicatescheck:DuplicatesCheck
        excelexport = psamm.commands.excelexport:ExcelExportCommand
        fastgapfill = psamm.commands.fastgapfill:FastGapFillCommand
        fba = psamm.commands.fba:FluxBalanceCommand
        fluxcheck = psamm.commands.fluxcheck:FluxConsistencyCommand
        fluxcoupling = psamm.commands.fluxcoupling:FluxCouplingCommand
        formulacheck = psamm.commands.formulacheck:FormulaBalanceCommand
        fva = psamm.commands.fva:FluxVariabilityCommand
        gapcheck = psamm.commands.gapcheck:GapCheckCommand
        gapfill = psamm.commands.gapfill:GapFillCommand
        genedelete = psamm.commands.genedelete:GeneDeletionCommand
        gimme = psamm.commands.gimme:GimmeCommand
        masscheck = psamm.commands.masscheck:MassConsistencyCommand
        primarypairs = psamm.commands.primarypairs:PrimaryPairsCommand
        randomsparse = psamm.commands.randomsparse:RandomSparseNetworkCommand
        robustness = psamm.commands.robustness:RobustnessCommand
        sbmlexport = psamm.commands.sbmlexport:SBMLExport
        search = psamm.commands.search:SearchCommand
        tableexport = psamm.commands.tableexport:ExportTableCommand
        psammotate = psamm.commands.psammotate:PsammotateCommand
        modelmapping = psamm.commands.model_mapping:ModelMappingCommand
        vis = psamm.commands.vis:VisualizationCommand
        tmfa = psamm.commands.tmfa:TMFACommand

        [psamm.importer]
        JSON = psamm.importers.cobrajson:Importer
        SBML = psamm.importers.sbml:NonstrictImporter
        SBML-strict = psamm.importers.sbml:StrictImporter
        MATLAB = psamm.importers.matlab:Importer
    ''',

    test_suite='psamm.tests',

    install_requires=[
        'pyyaml>=4.2b1',
        'six',
        'xlsxwriter',
        'numpy',
        'scipy',
        'future',
        'pandas'
    ],
    extras_require={
        'docs': ['sphinx', 'sphinx_rtd_theme', 'mock'],
        ':python_version=="2.7"': ['enum34'],
        ':python_version=="3.3"': ['enum34']
    })



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Ke Zhang <kzhang@my.uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import division

import logging
from itertools import product
from collections import Counter

from six import iteritems
from six.moves import reduce

from .formula import Formula, Atom

from fractions import gcd

logger = logging.getLogger(__name__)


def element_weight(element):
    """Return default element weight.

    This is the default element weight function.
    """
    if element == Atom.H:
        return 0.0
    elif element == Atom.C:
        return 1.0
    return 0.82


def _jaccard_similarity(f1, f2, weight_func):
    """Calculate generalized Jaccard similarity of formulas.

    Returns the weighted similarity value or None if there is no overlap
    at all. If the union of the formulas has a weight of zero (i.e. the
    denominator in the Jaccard similarity is zero), a value of zero is
    returned.
    """
    elements = set(f1)
    elements.update(f2)

    count, w_count, w_total = 0, 0, 0
    for element in elements:
        mi = min(f1.get(element, 0), f2.get(element, 0))
        mx = max(f1.get(element, 0), f2.get(element, 0))
        count += mi
        w = weight_func(element)
        w_count += w * mi
        w_total += w * mx

    if count == 0:
        return None

    return 0.0 if w_total == 0.0 else w_count / w_total


def _reaction_to_dicts(reaction):
    """Convert a reaction to reduced left, right dictionaries.

    Returns a pair of (left, right) dictionaries mapping compounds to
    normalized integer stoichiometric values. If a compound occurs multiple
    times on one side, the occurences are combined into a single entry in the
    dictionary.
    """
    def dict_from_iter_sum(it, div):
        d = {}
        for k, v in it:
            if k not in d:
                d[k] = 0
            d[k] += int(v / div)
        return d

    div = reduce(gcd, (abs(v) for _, v in reaction.compounds), 0)
    if div == 0:
        raise ValueError('Empty reaction')

    left = dict_from_iter_sum(reaction.left, div)
    right = dict_from_iter_sum(reaction.right, div)

    return left, right


class _CompoundInstance(object):
    """Instance of a compound in a reaction during matching.

    A compound in a reaction has a number of instances corresponding to the
    stoichiometric value. The instance is defined by the compound and index
    number. The compound instance also has an attached formula which can be
    updated through the ``formula`` property. The compound instance can be
    created as a copy of an existing instance, or from a
    :psamm:`psamm.reaction.Compound`, number, and
    :psamm:`psamm.formula.Formula`.
    """
    def __init__(self, *args):
        if len(args) == 1 and isinstance(args[0], _CompoundInstance):
            self._compound = args[0]._compound
            self._index = args[0]._index
            self._formula = args[0]._formula
        elif len(args) == 3:
            self._compound = args[0]
            self._index = args[1]
            self._formula = args[2]
        else:
            raise ValueError('Invalid arguments')

    @property
    def compound(self):
        """Return the :class:`psamm.reaction.Compound`."""
        return self._compound

    @property
    def index(self):
        """Return the index."""
        return self._index

    @property
    def formula(self):
        """Return the formula."""
        return self._formula

    @formula.setter
    def formula(self, value):
        self._formula = value

    def __repr__(self):
        return str('<{} {} of {!r}>').format(
            self.__class__.__name__, self._index, self._compound)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self._compound == other._index and
                self._index == other._index)

    def __hash__(self):
        return hash((self._compound, self._index))


def predict_compound_pairs_iterated(
        reactions, formulas, ambiguous=False, prior=(1, 43),
        max_iterations=None, element_weight=element_weight):
    """Predict reaction pairs using iterated method.

    Returns a tuple containing a dictionary of predictions keyed by the
    reaction IDs, and the final number of iterations. Each reaction prediction
    entry contains a tuple with a dictionary of transfers and a dictionary of
    unbalanced compounds. The dictionary of unbalanced compounds is empty only
    if the reaction is balanced.

    Args:
        reactions: Dictionary or pair-iterable of (id, equation) pairs.
            IDs must be any hashable reaction identifier (e.g. string) and
            equation must be :class:`psamm.reaction.Reaction` objects.
        ambiguous: True or False value to indicate if the ambiguous
            reactions should be printed in the debugging output.
        formulas: Dictionary mapping compound IDs to
            :class:`psamm.formula.Formula`. Formulas must be flattened.
        prior: Tuple of (alpha, beta) parameters for the MAP inference.
            If not provided, the default parameters will be used: (1, 43).
        max_iterations: Maximum iterations to run before stopping. If the
            stopping condition is reached before this number of iterations,
            the procedure also stops. If None, the procedure only stops when
            the stopping condition is reached.
        element_weight: A function providing returning weight value for the
            given :class:`psamm.formula.Atom` or
            :class:`psamm.formula.Radical`. If not provided, the default weight
            will be used (H=0, C=1, *=0.82)
    """
    prior_alpha, prior_beta = prior
    reactions = dict(reactions)

    pair_reactions = {}
    possible_pairs = Counter()
    tie_breakers = {}
    for reaction_id, equation in iteritems(reactions):
        tie_breakers[reaction_id] = {}
        for (c1, _), (c2, _) in product(equation.left, equation.right):
            spair = tuple(sorted([c1.name, c2.name]))
            possible_pairs[spair] += 1
            pair_reactions.setdefault(spair, set()).add(reaction_id)

    final_ambiguous_reactions = {}

    def print_ambiguous_summary():
        '''Function for summarizing and printing ambiguous compound pairs

        '''
        final_ambiguous_count = 0
        final_ambiguous_hydrogen = 0
        for reaction, pairs in sorted(iteritems(final_ambiguous_reactions)):
            all_hydrogen = False
            if len(pairs) > 0:
                final_ambiguous_count += 1
                all_hydrogen = True

                for pair in pairs:
                    for _, _, formula in pair:
                        df = dict(formula.items())
                        if len(df) > 1 or Atom.H not in df:
                            all_hydrogen = False

                if all_hydrogen:
                    final_ambiguous_hydrogen += 1
                    logger.info('Ambiguous Hydrogen '
                                'Transfers in Reaction: {}'.format(reaction))
                else:
                    logger.info('Ambiguous Non-Hydrogen '
                                'Transfers in Reaction: {}'.format(reaction))

        logger.info('{} reactions were decided with ambiguity'.format(
            final_ambiguous_count))
        logger.info('Only hydrogen in decision: {} vs non hydrogen: {}'.format(
            final_ambiguous_hydrogen,
            final_ambiguous_count - final_ambiguous_hydrogen))

    next_reactions = set(reactions)
    pairs_predicted = None
    prediction = {}
    weights = {}
    iteration = 0
    while len(next_reactions) > 0:
        iteration += 1
        if max_iterations is not None and iteration > max_iterations:
            break

        logger.info('Iteration {}: {} reactions...'.format(
            iteration, len(next_reactions)))

        for reaction_id in next_reactions:
            result = predict_compound_pairs(
                reactions[reaction_id], formulas, weights, element_weight)
            if result is None:
                continue

            transfer, balance, ambiguous_pairs = result
            final_ambiguous_reactions[reaction_id] = ambiguous_pairs

            rpairs = {}
            for ((c1, _), (c2, _)), form in iteritems(transfer):
                rpairs.setdefault((c1, c2), []).append(form)

            prediction[reaction_id] = rpairs, balance
        if ambiguous is True:
            logger.info('Ambiguous Reactions:')
            print_ambiguous_summary()

        pairs_predicted = Counter()
        for reaction_id, (rpairs, _) in iteritems(prediction):
            for c1, c2 in rpairs:
                spair = tuple(sorted([c1.name, c2.name]))
                pairs_predicted[spair] += 1

        next_reactions = set()
        for spair, total in sorted(iteritems(possible_pairs)):
            pred = pairs_predicted[spair]

            # The weight is set to the maximum a posteriori (MAP) estimate
            # of the primary pair probability distribution.
            posterior_alpha = prior_alpha + pred
            posterior_beta = prior_beta + total - pred
            pair_weight = ((posterior_alpha - 1) /
                           (posterior_alpha + posterior_beta - 2))

            if (spair not in weights or
                    abs(pair_weight - weights[spair]) > 1e-5):
                next_reactions.update(pair_reactions[spair])

            c1, c2 = spair
            weights[c1, c2] = pair_weight
            weights[c2, c1] = pair_weight

    return prediction, iteration


def _match_greedily(reaction, compound_formula, score_func):
    """Match compounds greedily based on score function.

    Args:
        reaction: Reaction equation :class:`psamm.reaction.Reaction`.
        compound_formula: Dictionary mapping compound IDs to
            :class:`psamm.formula.Formula`. Formulas must be flattened.
        score_func: Function that takes two :class:`_CompoundInstance` and
            returns the score.
    """
    uninstantiated_left, uninstantiated_right = _reaction_to_dicts(reaction)
    ambiguous_pairs = set()

    def compound_instances(uninstantiated):
        instances = []
        for compound, value in iteritems(uninstantiated):
            if value > 0:
                f = compound_formula[compound.name]
                instances.append(_CompoundInstance(compound, value, f))

        for inst in instances:
            uninstantiated[inst.compound] -= 1

        return instances

    def instantiate(uninstantiated, compound):
        n = uninstantiated[compound]
        if n > 0:
            f = compound_formula[compound.name]
            inst = _CompoundInstance(compound, n, f)
            uninstantiated[compound] -= 1
            return inst

        return None

    left = compound_instances(uninstantiated_left)
    right = compound_instances(uninstantiated_right)
    instances = left + right

    pairs = {}
    for inst1, inst2 in product(left, right):
        result = score_func(inst1, inst2)
        if result is not None:
            pairs[inst1, inst2] = result

    def inst_pair_sort_key(entry):
        """Sort key for finding best match among instance pairs.

        Rank by score in general but always match identical compounds first
        (these will always have score equal to one but are handled specially
        to put them ahead of other compounds with score equal to one). Use
        compound names to break ties to produce a deterministic result.
        """
        (inst1, inst2), score = entry
        c1, c2 = inst1.compound, inst2.compound
        same_compound = c1.name == c2.name and c1.compartment != c2.compartment
        return same_compound, score, c1.name, c2.name

    transfer = {}
    while len(pairs) > 0:
        (inst1, inst2), _ = max(iteritems(pairs), key=inst_pair_sort_key)
        common = inst1.formula & inst2.formula
        sorted_pairs = sorted(iteritems(pairs),
                              key=inst_pair_sort_key, reverse=True)
        max_sort_key = inst_pair_sort_key(sorted_pairs[0])

        ambiguous_entry = {(inst1.compound, inst2.compound, common)}
        for (other_inst1, other_inst2), other_score in sorted_pairs[1:]:
            other_sort_key = inst_pair_sort_key(
                ((other_inst1, other_inst2), other_score))
            if other_sort_key[:2] < max_sort_key[:2]:
                break

            other_common = other_inst1.formula & other_inst2.formula
            ambiguous_entry.add(
                (other_inst1.compound, other_inst2.compound, other_common))

        if len(ambiguous_entry) > 1:
            ambiguous_pairs.add(tuple(ambiguous_entry))

        key = (inst1.compound, inst1.index), (inst2.compound, inst2.index)
        if key not in transfer:
            transfer[key] = Formula()
        transfer[key] |= common

        for inst in (inst1, inst2):
            inst.formula -= common

        to_insert = set()

        inst = instantiate(uninstantiated_left, inst1.compound)
        if inst is not None:
            left.append(inst)
            instances.append(inst)
            to_insert.add(inst)

        inst = instantiate(uninstantiated_right, inst2.compound)
        if inst is not None:
            right.append(inst)
            instances.append(inst)
            to_insert.add(inst)

        to_update = {inst1, inst2}

        to_delete = set()
        for inst1, inst2 in pairs:
            if inst1 in to_update or inst2 in to_update:
                if len(inst1.formula) > 0 and len(inst2.formula) > 0:
                    result = score_func(inst1, inst2)
                    if result is None:
                        to_delete.add((inst1, inst2))
                    else:
                        pairs[inst1, inst2] = result
                else:
                    to_delete.add((inst1, inst2))

        for pair in to_delete:
            del pairs[pair]

        for inst1, inst2 in product(left, right):
            if inst1 in to_insert or inst2 in to_insert:
                result = score_func(inst1, inst2)
                if result is not None:
                    pairs[inst1, inst2] = result

    balance = {}
    for inst in instances:
        if len(inst.formula) > 0:
            key = inst.compound, inst.index
            balance[key] = inst.formula

    return transfer, balance, ambiguous_pairs


def predict_compound_pairs(reaction, compound_formula, pair_weights={},
                           weight_func=element_weight):
    """Predict compound pairs for a single reaction.

    Performs greedy matching on reaction compounds using a scoring function
    that uses generalized Jaccard similarity corrected by the weights in the
    given dictionary. Returns a tuple of a transfer dictionary and a dictionary
    of unbalanced compounds. The dictionary of unbalanced compounds is empty
    only if the reaction is balanced.

    Args:
        reaction: :class:`psamm.reaction.Reaction`.
        compound_formula: Dictionary mapping compound IDs to
            :class:`psamm.formula.Formula`. Formulas must be flattened.
        pair_weights: Dictionary mapping pairs of compound IDs to correction
            values. This value is multiplied by the calculated Jaccard
            similarity. If a pair is not in the dictionary, the value 1 is
            used. Pairs are looked up in the weights dictionary as a tuple of
            compound names (``c1``, ``c2``) where ``c1`` is the left-hand side
            and ``c2`` is the right-hand side.
        weight_func: Weight function for caclulating the generalized Jaccard
            similarity. This function will be given an
            :class:`psamm.formula.Atom` or :class:`psamm.formula.Radical` and
            should return a corresponding weight.
    """
    def score_func(inst1, inst2):
        score = _jaccard_similarity(
            inst1.formula, inst2.formula, weight_func)
        if score is None:
            return None
        pair = inst1.compound.name, inst2.compound.name
        pair_weight = pair_weights.get(pair, 1.0)
        return pair_weight * score

    return _match_greedily(reaction, compound_formula, score_func)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import math
import logging

from ..command import Command, FilePrefixAppendAction, convert_to_unicode
from ..balancecheck import charge_balance

logger = logging.getLogger(__name__)


class ChargeBalanceCommand(Command):
    """Check whether compound charge is balanced.

    Balanced reactions are those reactions where the total charge
    is consistent on the left and right side of the reaction equation.
    Reactions that are not balanced will be printed out.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--exclude', metavar='reaction', type=convert_to_unicode,
            action=FilePrefixAppendAction,
            default=[], help='Exclude reaction from balance check')
        parser.add_argument(
            '--epsilon', metavar='epsilon', type=float, default=1e-6,
            help='Threshold for charge imbalance to be considered zero'
        )
        super(ChargeBalanceCommand, cls).init_parser(parser)

    def run(self):
        """Run charge balance command"""

        # Load compound information
        def compound_name(id):
            if id not in self._model.compounds:
                return id
            return self._model.compounds[id].properties.get('name', id)

        # Create a set of excluded reactions
        exclude = set(self._args.exclude)
        count = 0
        unbalanced = 0
        unchecked = 0
        for reaction, charge in charge_balance(self._model):
            count += 1

            if reaction.id in exclude or reaction.equation is None:
                continue

            if math.isnan(charge):
                logger.debug('Not checking reaction {};'
                             ' missing charge'.format(reaction.id))
                unchecked += 1
            elif abs(charge) > self._args.epsilon:
                unbalanced += 1
                rxt = reaction.equation.translated_compounds(compound_name)
                print('{}\t{}\t{}'.format(reaction.id, charge, rxt))

        logger.info('Unbalanced reactions: {}/{}'.format(unbalanced, count))
        logger.info('Unchecked reactions due to missing charge: {}/{}'.format(
            unchecked, count))
        logger.info('Reactions excluded from check: {}/{}'.format(
            len(exclude), count))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2016-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016  Chao liu <lcddzyx@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest

from psamm.datasource import native
from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm.datasource.reaction import parse_reaction
from psamm import gapfilling


class TestAddReactions(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) A[c]'))
        self.database.set_reaction('rxn_2', parse_reaction('A[c] <=> B[c]'))
        self.database.set_reaction('rxn_3', parse_reaction('A[c] => D[e]'))
        self.database.set_reaction('rxn_4', parse_reaction('A[c] => C[c]'))
        self.database.set_reaction('rxn_5', parse_reaction('C[c] => D[e]'))
        self.database.set_reaction('rxn_6', parse_reaction('D[e] =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)

    def test_add_all_database_reactions(self):
        # Should get added
        self.database.set_reaction('rxn_7', parse_reaction('D[c] => E[c]'))
        # Not added because of compartment
        self.database.set_reaction('rxn_8', parse_reaction('D[c] => E[p]'))
        added = gapfilling.add_all_database_reactions(self.model, {'c', 'e'})
        self.assertEqual(added, {'rxn_7'})
        self.assertEqual(set(self.model.reactions), {
            'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6', 'rxn_7'
        })

    def test_add_all_database_reactions_none(self):
        added = gapfilling.add_all_database_reactions(self.model, {'c', 'e'})
        self.assertEqual(added, set())
        self.assertEqual(set(self.model.reactions), {
            'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4', 'rxn_5', 'rxn_6'})

    def test_add_all_transport_reactions(self):
        added = gapfilling.add_all_transport_reactions(
            self.model, {('e', 'c')})
        for reaction in added:
            compartments = tuple(c.compartment for c, _ in
                                 self.model.get_reaction_values(reaction))
            self.assertEqual(len(compartments), 2)
            self.assertTrue('e' in compartments)
            self.assertTrue(compartments[0] != compartments[1])


class TestCreateExtendedModel(unittest.TestCase):
    def setUp(self):
        reader = native.ModelReader({
            'compartments': [
                {
                    'id': 'c',
                    'adjacent_to': 'e'
                }, {
                    'id': 'e'
                }
            ],
            'reactions': [
                {
                    'id': 'rxn_1',
                    'equation': 'A[e] <=> B[c]'
                }, {
                    'id': 'rxn_2',
                    'equation': 'A[e] => C[c]'
                }, {
                    'id': 'rxn_3',
                    'equation': 'A[e] => D[e]'
                }, {
                    'id': 'rxn_4',
                    'equation': 'C[c] => D[e]'
                }
            ],
            'compounds': [
                {'id': 'A'},
                {'id': 'B'},
                {'id': 'C'},
                {'id': 'D'}
            ],
            'exchange': [{
                'compartment': 'e',
                'compounds': [
                    {
                        'id': 'A',
                        'reaction': 'rxn_5'
                    },
                    {
                        'id': 'D',
                        'reaction': 'rxn_6'
                    }
                ]
            }],
            'model': [{
                'reactions': [
                    'rxn_1',
                    'rxn_2'
                ]
            }]
        })
        self._model = reader.create_model()

    def test_create_model_extended(self):
        expected_reactions = set([
            'rxn_1',
            'rxn_2',
            'rxn_3',
            'rxn_4',
            'rxn_5',
            'rxn_6',
            'TP_A[c]_A[e]',
            'TP_B[c]_B[e]',
            'TP_C[c]_C[e]',
            'TP_D[c]_D[e]',
            'EX_B[e]',
            'EX_C[e]',
        ])

        expected_weights = {
            'rxn_3': 5.6,
            'rxn_4': 1.0,
            'TP_A[c]_A[e]': 3.0,
            'TP_B[c]_B[e]': 3.0,
            'TP_C[c]_C[e]': 3.0,
            'TP_D[c]_D[e]': 3.0,
            'EX_B[e]': 2.0,
            'EX_C[e]': 2.0
        }
        penalties = {'rxn_3': 5.6}

        model_extended, weights = gapfilling.create_extended_model(
            self._model,
            db_penalty=1.0,
            ex_penalty=2.0,
            tp_penalty=3.0,
            penalties=penalties)
        self.assertEqual(set(model_extended.reactions), expected_reactions)
        self.assertEqual(weights, expected_weights)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Jing Wang <wjingsjtu@gmail.com>

"""Calculate model mapping likelihood with bayesian."""
from __future__ import print_function, division
from future.utils import itervalues

from builtins import object, range
from itertools import product
from multiprocessing import Pool
import operator
import sys
from collections import namedtuple, defaultdict
import pandas as pd
import time

import psamm.bayesian_util as util
from psamm.formula import Formula
from psamm.expression.boolean import Expression
from functools import reduce

import logging
logger = logging.getLogger(__name__)


CompoundEntry = namedtuple(
    'CompoundEntry',
    ['id', 'name', 'formula', 'charge', 'kegg', 'cas'])

ReactionEntry = namedtuple(
    'ReactionEntry',
    ['id', 'name', 'genes', 'equation', 'subsystem', 'ec'])


class MappingModel(object):
    """Generate the internal structure for model mapping.

    Args:
        model: :class:`psamm.datasource.native.NativeModel`
    """

    def __init__(self, model):
        self._name = model.name
        self._read_compounds(model.compounds)
        self._read_reactions(model.reactions)

        self._genes = set()
        self._has_gene = list()
        for r in itervalues(self._reactions):
            if r.genes is not None:
                self._has_gene.append(r.id)
                e = Expression(r.genes)
                self._genes.update([g.symbol for g in e.variables])

    def _read_compounds(self, it):
        self._compounds = {}
        for compound in it:
            formula = getattr(compound, 'formula', None)
            if formula is not None:
                formula = Formula.parse(formula)
            else:
                formula = None

            if 'kegg' in compound.properties:
                compound_kegg = compound.properties['kegg']
            else:
                compound_kegg = None

            compound_id = compound.id
            entry = CompoundEntry(
                id=compound_id, name=getattr(compound, 'name', None),
                formula=formula, charge=getattr(compound, 'charge', None),
                kegg=compound_kegg,
                cas=getattr(compound, 'cas', None))
            self._compounds[compound_id] = entry

    def _read_reactions(self, it):
        self._reactions = {}
        for reaction in it:
            genes = getattr(reaction, 'genes', None)
            reaction_id = reaction.id
            equation = getattr(reaction, 'equation', None)
            entry = ReactionEntry(
                id=reaction_id, name=getattr(reaction, 'name', None),
                genes=genes, equation=equation,
                subsystem=getattr(reaction, 'subsystem', None),
                ec=getattr(reaction, 'ec', None))
            self._reactions[reaction_id] = entry

    @property
    def name(self):
        return self._name

    @property
    def reactions(self):
        return self._reactions

    @property
    def compounds(self):
        return self._compounds

    @property
    def genes(self):
        return self._genes

    def print_summary(self):
        """Print model summary"""
        print('Model: {}'.format(self.name))
        print('- Compounds: {}'.format(len(self.compounds)))
        print('- Reactions: {}'.format(len(self.reactions)))
        print('- Genes: {}'.format(len(self.genes)))
        print('- Reactions with gene association: {}'.format(
            len(self._has_gene)))

    def check_reaction_compounds(self):
        """Check that reaction compounds are defined in the model"""
        print('Checking model: {}'.format(self.name))
        consistent = True
        undefined_cpds = []
        for reaction in itervalues(self.reactions):
            if reaction.equation is not None:
                for compound, value in reaction.equation.compounds:
                    if compound.name not in self.compounds:
                        undefined_cpds.append(compound.name)
                        logger.error((
                            '{} in reaction {} is not defined in compound '
                            'list (such as compounds.yaml), please define it '
                            'in compound list before running modelmapping'
                        ).format(compound.name, reaction.id))
                        consistent = False
        return consistent


class BayesianCompoundPredictor(object):
    """Predict model compound mappings based on a Bayesian model.

    Args:
        model1: :class:`psamm.bayesian.MappingModel`.
        model2: :class:`psamm.bayesian.MappingModel`.
        nproc: number of processes used for mapping.
        outpath: the path to output the detailed log file.
        log: whether to output log file of the p_match and p_no_match.
        kegg: whether to compare the KEGG id.
    """

    def __init__(self, model1, model2, nproc=1,
                 outpath='.', log=False, kegg=False):
        self._model1 = model1
        self._model2 = model2
        self._column_list = [
            'p', 'p_id', 'p_name', 'p_charge', 'p_formula', 'p_kegg']
        self._compound_map_p = map_model_compounds(
            self._model1, self._model2, nproc, outpath, log=log, kegg=kegg)

    @property
    def model1(self):
        return self._model1

    @property
    def model2(self):
        return self._model2

    def map(self, c1, c2):
        return self._compound_map_p[0][c1, c2]

    def get_raw_map(self):
        """Return pandas.DataFrame style of raw mapping table."""
        compound_result = pd.DataFrame({
            self._column_list[i]: self._compound_map_p[i]
            for i in range(len(self._column_list))
        })
        return compound_result

    def get_best_map(self, threshold_compound=0):
        """Return :class:`pandas.DataFrame` style of best mapping for each
        query."""
        raw_map = self.get_raw_map()
        query = [i for i, j in raw_map.index.values]
        # use rank instead of idxmax to output multiple top-hitts
        best_index = raw_map.iloc[:, 0].groupby(query).rank(
            method='min', ascending=False).loc[lambda x: x == 1].index.values
        compound_best = raw_map.loc[best_index]
        compound_best.query(
            'p >= @threshold_compound', inplace=True)
        return compound_best

    def get_cpd_pred(self, threshold_compound=0):
        """Return the cpd_pred used for reaction mapping."""
        return self.get_best_map(threshold_compound).loc[:, 'p']


class BayesianReactionPredictor(object):
    """Predict model reaction mappings based on a Bayesian model.

    Args:
        model1: :class:`psamm.bayesian.MappingModel`.
        model2: :class:`psamm.bayesian.MappingModel`.
        cpd_pred: :class:`pandas.Series` with compound pairs as index,
                  compound mapping score as value.
        nproc: number of processes used for mapping.
        outpath: the path to output the detailed log file.
        log: whether to output log file of the p_match and p_no_match.
        gene: whether to compare the gene association.
        compartment_map: dictionary mapping compartment id in the query model
                         to the id in the target model.
        gene_map: dictionary mapping gene id in the query model to the id in
                  the target model.
    """

    def __init__(self, model1, model2, cpd_pred, nproc=1,
                 outpath='.', log=False, gene=False,
                 compartment_map={}, gene_map={}):
        self._model1 = model1
        self._model2 = model2
        self._parse_cpd_pred(cpd_pred)
        self._column_list = ['p', 'p_id', 'p_name', 'p_equation', 'p_genes']
        gene_map = self._reversible_map(gene_map)
        self._reaction_map_p = map_model_reactions(
            self._model1, self._model2, self._cpd_map, self._cpd_score, nproc,
            outpath, log=log, gene=gene, compartment_map=compartment_map,
            gene_map=gene_map)

    @property
    def model1(self):
        return self._model1

    @property
    def model2(self):
        return self._model2

    def map(self, r1, r2):
        return self._reaction_map_p[0][r1, r2]

    def _reversible_map(self, genemap):
        newmap = {}
        for k, v in genemap.items():
            newmap[k] = v
            newmap[v] = k
        return newmap

    def _parse_cpd_pred(self, cpd_pred):
        self._cpd_map = defaultdict(set)
        self._cpd_score = dict()
        for pair, score in cpd_pred.items():
            self._cpd_map[pair[0]].add(pair[1])
            self._cpd_score[pair[0]] = score

    def get_raw_map(self):
        """Return pandas.DataFrame style of raw mapping table."""
        reaction_result = pd.DataFrame({
            self._column_list[i]: self._reaction_map_p[i]
            for i in range(len(self._column_list))
        })
        return reaction_result

    def get_best_map(self, threshold_reaction=0):
        """Return pandas.DataFrame style of best mapping for each query."""
        raw_map = self.get_raw_map()
        query = [i for i, j in raw_map.index.values]
        # use rank instead of idxmax to output multiple top-hitts
        best_index = raw_map.iloc[:, 0].groupby(query).rank(
            method='min', ascending=False).loc[lambda x: x == 1].index.values
        reaction_best = raw_map.loc[best_index]
        reaction_best.query(
            'p >= @threshold_reaction', inplace=True)
        return reaction_best


def compound_id_likelihood(c1, c2, compound_prior, compound_id_marg):
    if util.id_equals(c1.id, c2.id):
        p_match = 0.4
        p_marg = compound_id_marg
        p_no_match = max(
            0, (p_marg - p_match * compound_prior) / (1.0 - compound_prior))
    else:
        p_match = 0.6
        p_marg = 1.0 - compound_id_marg
        p_no_match = max(
            0, (p_marg - p_match * compound_prior) / (1.0 - compound_prior))

    return p_match, p_no_match


def compound_name_likelihood(c1, c2, compound_prior, compound_name_marg):
    if util.name_equals(c1.name, c2.name):
        p_match = 0.60
        p_marg = compound_name_marg
        p_no_match = max(
            0, (p_marg - p_match * compound_prior) / (1.0 - compound_prior))
    else:
        p_match = 0.40
        p_marg = 1.0 - compound_name_marg
        p_no_match = max(
            0, (p_marg - p_match * compound_prior) / (1.0 - compound_prior))

    return p_match, p_no_match  # , p_marg


def compound_charge_likelihood(
        c1, c2, compound_prior,
        compound_charge_equal_marg, compound_charge_not_equal_marg):
    if c1.charge is None or c2.charge is None:
        # p value of observing undefined charge
        # it is independent of the condition of match or not
        p_match = 1
        p_no_match = 1
    elif c1.charge == c2.charge:
        p_match = 0.9
        p_no_match = max(
            0,
            ((compound_charge_equal_marg - p_match * compound_prior) /
             (1.0 - compound_prior)))
    else:
        p_match = 0.1
        p_no_match = max(
            0,
            ((compound_charge_not_equal_marg - p_match * compound_prior) /
             (1.0 - compound_prior)))

    return p_match, p_no_match


def compound_formula_likelihood(
        c1, c2, compound_prior,
        compound_formula_equal_marg, compound_formula_not_equal_marg):
    if c1.formula is None or c2.formula is None:
        # p value of observing undefined formula
        # it is independent of the condition of match or not
        p_match = 1
        p_no_match = 1
    elif util.formula_equals(c1.formula, c2.formula, c1.charge, c2.charge):
        p_match = 0.9
        p_no_match = max(
            0,
            ((compound_formula_equal_marg - p_match * compound_prior) /
             (1.0 - compound_prior)))
    else:
        p_match = 0.1
        p_no_match = max(
            0,
            ((compound_formula_not_equal_marg - p_match * compound_prior) /
             (1.0 - compound_prior)))

    return p_match, p_no_match


def compound_kegg_likelihood(
        c1, c2, compound_prior,
        compound_kegg_equal_marg, compound_kegg_not_equal_marg):
    if c1.kegg is None or c2.kegg is None:
        # p value of observing undefined KEGG
        # it is independent of the condition of match or not
        p_match = 1
        p_no_match = 1
    elif c1.kegg == c2.kegg:
        p_match = 0.65
        p_no_match = max(
            0,
            ((compound_kegg_equal_marg - p_match * compound_prior) /
             (1.0 - compound_prior)))
    else:
        p_match = 0.35
        p_no_match = max(
            0,
            ((compound_kegg_not_equal_marg - p_match * compound_prior) /
             (1.0 - compound_prior)))

    return p_match, p_no_match


def reaction_id_likelihood(
        r1, r2, reaction_prior,
        reaction_id_equal_marg, reaction_id_not_equal_marg):
    if util.id_equals(r1.id, r2.id):
        p_match = 0.3
        p_no_match = max(
            0,
            ((reaction_id_equal_marg - p_match * reaction_prior) /
             (1.0 - reaction_prior)))
    else:
        p_match = 0.7
        p_no_match = max(
            0,
            ((reaction_id_not_equal_marg - p_match * reaction_prior) /
             (1.0 - reaction_prior)))

    return p_match, p_no_match


def reaction_name_likelihood(r1, r2, reaction_prior, reaction_name_marg):
    if util.name_equals(r1.name, r2.name):
        p_match = 0.2
        p_no_match = max(
            0,
            ((reaction_name_marg - p_match * reaction_prior) /
             (1.0 - reaction_prior)))
    else:
        p_match = 0.8
        p_no_match = max(
            0,
            ((1.0 - reaction_name_marg - p_match * reaction_prior) /
             (1.0 - reaction_prior)))

    return p_match, p_no_match


def reaction_equation_mapping_approx_max_likelihood(
        cpd_set1, cpd_set2, cpd_map, cpd_score, compartment_map={}):
    """Calculate equation likelihood based on compound mapping."""
    p_match = 1.0
    p_no_match = 1.0

    compartment_dict = defaultdict(set)
    for c in cpd_set2:
        compartment_dict[c.name].add(c.compartment)

    # get the possible best-match pairs, score as the key
    best_match = defaultdict(set)
    for c1 in cpd_set1:
        for c2 in cpd_map[c1.name].intersection(compartment_dict.keys()):
            # the compound pair should be in the same compartment
            if (compartment_map.get(c1.compartment, c1.compartment)
                    in compartment_dict[c2]):
                best_match[cpd_score[c1.name]].add((c1.name, c2))

    # translate each element to compound id
    cpd_set1 = set(c.name for c in cpd_set1)
    cpd_set2 = set(c.name for c in cpd_set2)

    # sort by scores, get the most possible compound pairs first
    for score in sorted(best_match.keys(), reverse=True):
        for c1, c2 in best_match[score]:
            if (c1 in cpd_set1) and (c2 in cpd_set2):
                # the possibility that compounds are equal
                p_match *= score * 0.9 + (1 - score) * 0.1
                p_no_match *= score * 0.1 + (1 - score) * 0.9
                cpd_set1.remove(c1)
                cpd_set2.remove(c2)
                if len(cpd_set1) == 0 or len(cpd_set2) == 0:
                    break

    for c in cpd_set1:
        p_match *= 0.1
        p_no_match *= 0.9
    for c in cpd_set2:
        p_match *= 0.1
        p_no_match *= 0.9

    return p_match, p_no_match


def reaction_equation_compound_mapping_likelihood(
        r1, r2, *args, **kwargs):
    """Get the likelihood of reaction equations

    Args:
        r1, r2: two `RactionEntry` objects to be compared
    args, kwargs:
        cpd_map: dictionary mapping compound id in the query model to a set
                 of best-mapping compound ids in the target model.
        cpd_score: dictionary mapping compound id in the query model to
                   its best mapping score during compound mapping.
        compartment_map: dictionary mapping compartment id in the query model
                         to the id in the target model.
    """
    if r1.equation is None or r2.equation is None:
        # p value of observing undefined equation
        # it is independent of the condition of match or not
        p_match = 1
        p_no_match = 1
    else:
        p_match, p_no_match = get_best_p_value_set(r1, r2, *args, **kwargs)

    return p_match, p_no_match


def get_best_p_value_set(r1, r2, *args, **kwargs):
    """Assume equations may have reversed direction, report best mapping p."""
    cpd_set1_left = get_cpd_set(r1.equation, left=True)
    cpd_set1_right = get_cpd_set(r1.equation, left=False)
    cpd_set2_left = get_cpd_set(r2.equation, left=True)
    cpd_set2_right = get_cpd_set(r2.equation, left=False)

    # assume equations have the same direction
    p_forward_match, p_forward_no_match = merge_partial_p_set(
        cpd_set1_left, cpd_set2_left,
        cpd_set1_right, cpd_set2_right, *args, **kwargs)
    # assume equations have the reversed direction
    p_reverse_match, p_reverse_no_match = merge_partial_p_set(
        cpd_set1_left, cpd_set2_right,
        cpd_set1_right, cpd_set2_left, *args, **kwargs)

    # maintain the direction with better p values
    if (p_forward_match / p_forward_no_match >=
            p_reverse_match / p_reverse_no_match):
        p_match = p_forward_match
        p_no_match = p_forward_no_match
    else:
        p_match = p_reverse_match
        p_no_match = p_reverse_no_match

    return p_match, p_no_match


def merge_partial_p_set(cpd_set1_left, cpd_set2_left,
                        cpd_set1_right, cpd_set2_right, *args, **kwargs):
    """Merge the left hand side and right hand side p values together.

    The compound mapping is done separately on left hand side and
    right hand side.
    Then the corresponding p_match and p_no_match are merged together.
    """
    p_set_left = \
        reaction_equation_mapping_approx_max_likelihood(
            cpd_set1_left, cpd_set2_left, *args, **kwargs)
    p_set_right = \
        reaction_equation_mapping_approx_max_likelihood(
            cpd_set1_right, cpd_set2_right, *args, **kwargs)
    p_match = p_set_left[0] * p_set_right[0]
    p_no_match = p_set_left[1] * p_set_right[1]
    return p_match, p_no_match


def get_cpd_set(equation, left=True):
    if left:  # value of left-side compound is negtive
        coef = 1
    else:  # value of right-side compound is positive
        coef = -1
    # pick compounds at one side only
    cpd_set = set(
        compound
        for compound, value in equation.compounds
        if coef * value < 0)
    return cpd_set


def reaction_genes_likelihood(r1, r2, reaction_prior, reaction_genes_marg,
                              reaction_genes_not_equal_marg, gene_map={}):
    if r1.genes is None or r2.genes is None:
        p_match = 1
        p_no_match = 1
    elif util.genes_equals(r1.genes, r2.genes, gene_map):
        p_match = 0.2
        p_no_match = max(
            0,
            ((reaction_genes_marg - p_match * reaction_prior) /
             (1.0 - reaction_prior)))
    else:
        p_match = 0.8
        p_no_match = max(
            0,
            ((reaction_genes_not_equal_marg - p_match * reaction_prior) /
             (1.0 - reaction_prior)))

    return p_match, p_no_match


def fake_likelihood(e1, e2):
    """Generate fake likelihood if corresponding mapping is not required."""
    return 1, 1


def generate_likelihood(tasks):
    pair, likelihood, args, kwargs = tasks
    e1, e2 = pair
    p1, p2 = likelihood(e1, e2, *args, **kwargs)
    return e1.id, e2.id, p1, p2


def pairwise_likelihood(pool, chunksize, model1, model2, likelihood,
                        *args, **kwargs):
    """Compute likelihood of all pairwise comparisons.

    Returns likelihoods as a dataframe with a column for each hypothesis.
    """
    tasks = (((e1, e2), likelihood, args, kwargs)
             for e1, e2 in product(itervalues(model1), itervalues(model2)))
    result = pool.map(generate_likelihood, tasks, chunksize=chunksize)
    return pd.DataFrame.from_records(result, index=('e1', 'e2'),
                                     columns=('e1', 'e2', 'p1', 'p2'))


def likelihood_products(likelihood_dfs):
    """Combine likelihood dataframes."""
    return reduce(operator.mul, likelihood_dfs, 1.0)


def bayes_posterior(prior, likelihood_df):
    """Calculate posterior given likelihoods and prior."""
    p_1 = prior * likelihood_df.iloc[:, 0]
    p_2 = (1.0 - prior) * likelihood_df.iloc[:, 1]
    return p_1 / (p_1 + p_2)


def parallel_equel(tasks):
    func, params = tasks
    return func(*params)


def map_model_compounds(model1, model2, nproc=1, outpath='.',
                        log=False, kegg=False):
    """Map compounds of two models."""
    compound_pairs = len(model1.compounds) * len(model2.compounds)

    # Compound prior
    # For the prior, use a guesstimate that 95% of the
    # smaller model can be mapped.
    compound_prior = (0.95 * min(len(model1.compounds),
                                 len(model2.compounds))) / compound_pairs

    # Initialize parallel pool of workers
    chunksize = compound_pairs // nproc
    pool = Pool(nproc)

    t = time.time()
    # Compound ID
    print('Calculating compound ID likelihoods...', end=' ')
    sys.stdout.flush()

    # Marginal probability of observing two equal compound IDs
    tasks = ((util.id_equals, (c1.id, c2.id)) for c1, c2 in product(
        itervalues(model1.compounds), itervalues(model2.compounds)))
    result = pool.map(parallel_equel, tasks, chunksize=chunksize)
    compound_id_marg = sum(result) / float(compound_pairs)

    compound_id_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.compounds, model2.compounds,
        compound_id_likelihood, compound_prior, compound_id_marg)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()
    # Compound name
    print('Calculating compound name likelihoods...', end=' ')
    sys.stdout.flush()

    # Marginal probability of observing two similar names
    tasks = ((util.name_equals, (c1.name, c2.name)) for c1, c2 in product(
        itervalues(model1.compounds), itervalues(model2.compounds)))
    result = pool.map(parallel_equel, tasks, chunksize=chunksize)
    compound_name_marg = sum(result) / float(compound_pairs)

    compound_name_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.compounds, model2.compounds,
        compound_name_likelihood, compound_prior, compound_name_marg)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()

    # Compound charge
    print('Calculating compound charge likelihoods...', end=' ')
    sys.stdout.flush()

    # Marginal probability of observing two compounds with the same charge
    compound_charge_equal_marg = sum(
        c1.charge is not None and
        c2.charge is not None and
        c1.charge == c2.charge
        for c1, c2 in product(
            itervalues(model1.compounds), itervalues(model2.compounds))
    ) / compound_pairs

    # Marginal probability of observing two compounds with different charge
    compound_charge_not_equal_marg = sum(
        c1.charge is not None and
        c2.charge is not None and
        c1.charge != c2.charge
        for c1, c2 in product(
            itervalues(model1.compounds), itervalues(model2.compounds))
    ) / compound_pairs

    compound_charge_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.compounds, model2.compounds,
        compound_charge_likelihood,
        compound_prior,
        compound_charge_equal_marg,
        compound_charge_not_equal_marg)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()

    # Compound formula
    print('Calculating compound formula likelihoods...', end=' ')
    sys.stdout.flush()

    # Marginal probability of observing two compounds with the same formula
    tasks = ((
        util.formula_equals,
        (c1.formula, c2.formula, c1.charge, c2.charge))
        for c1, c2 in product(
            itervalues(model1.compounds), itervalues(model2.compounds)))
    result = pool.map(parallel_equel, tasks, chunksize=chunksize)
    compound_formula_equal_marg = sum(result) / float(compound_pairs)

    # Marginal probability of observing two compounds with different formula
    compound_formula_not_equal_marg = 1.0 - compound_formula_equal_marg - (
        sum(c1.formula is None or c2.formula is None
            for c1, c2 in product(itervalues(model1.compounds),
                                  itervalues(model2.compounds))) /
        compound_pairs)

    compound_formula_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.compounds, model2.compounds,
        compound_formula_likelihood,
        compound_prior, compound_formula_equal_marg,
        compound_formula_not_equal_marg)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()

    # Compound KEGG id
    if kegg:  # run KEGG id mapping
        print('Calculating compound KEGG ID likelihoods...', end=' ')
        sys.stdout.flush()

        # Marginal probability of observing two compounds
        # where KEGG ids are equal
        compound_kegg_equal_marg = sum(
            c1.kegg is not None and
            c2.kegg is not None and
            c1.kegg == c2.kegg
            for c1, c2 in product(
                itervalues(model1.compounds),
                itervalues(model2.compounds))
        ) / compound_pairs

        # Marginal probability of observing two compounds
        # where KEGG ids are different
        compound_kegg_not_equal_marg = sum(
            c1.kegg is not None and
            c2.kegg is not None and
            c1.kegg != c2.kegg for c1, c2 in product(
                itervalues(model1.compounds),
                itervalues(model2.compounds))
        ) / compound_pairs

        compound_kegg_likelihoods = pairwise_likelihood(
            pool, chunksize, model1.compounds, model2.compounds,
            compound_kegg_likelihood,
            compound_prior, compound_kegg_equal_marg,
            compound_kegg_not_equal_marg)

        print('%.2f seconds' % (time.time() - t))
        t = time.time()
    else:  # run fake mapping
        compound_kegg_likelihoods = pairwise_likelihood(
            pool, chunksize, model1.compounds, model2.compounds,
            fake_likelihood)

    pool.close()
    pool.join()

    if log:
        merge_result = pd.merge(compound_id_likelihoods,
                                compound_name_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_id', '_name'))
        merge_result = pd.merge(merge_result, compound_charge_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_name', '_charge'))
        merge_result = pd.merge(merge_result, compound_formula_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_charge', '_formula'))
        merge_result = pd.merge(merge_result, compound_kegg_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_formula', '_kegg'))

        merge_result.to_csv(outpath + '/compound_log.tsv', sep='\t')

    all_likelihoods = [compound_id_likelihoods,
                       compound_name_likelihoods,
                       compound_charge_likelihoods,
                       compound_formula_likelihoods,
                       compound_kegg_likelihoods]

    return (bayes_posterior(compound_prior,
                            likelihood_products(all_likelihoods)),
            bayes_posterior(compound_prior, compound_id_likelihoods),
            bayes_posterior(compound_prior, compound_name_likelihoods),
            bayes_posterior(compound_prior, compound_charge_likelihoods),
            bayes_posterior(compound_prior, compound_formula_likelihoods),
            bayes_posterior(compound_prior, compound_kegg_likelihoods))


def map_model_reactions(model1, model2, cpd_map, cpd_score, nproc=1,
                        outpath='.', log=False, gene=False, compartment_map={},
                        gene_map={}):
    """Map reactions of two models."""
    # Mapping of reactions
    reaction_pairs = len(model1.reactions) * len(model2.reactions)

    # Reaction prior
    # For the prior, use a guesstimate that 95%
    # of the smaller model can be mapped.
    reaction_prior = (0.95 * min(len(model1.reactions),
                                 len(model2.reactions))) / reaction_pairs

    # Initialize parallel pool of workers
    chunksize = reaction_pairs // nproc
    pool = Pool(nproc)

    # Reaction ID
    t = time.time()
    print('Calculating reaction ID likelihoods...', end=' ')
    sys.stdout.flush()

    # Marginal probability of observing two reactions with the same ids.
    tasks = ((util.id_equals, (r1.id, r2.id)) for r1, r2 in product(
        itervalues(model1.reactions),
        itervalues(model2.reactions)))
    result = pool.map(parallel_equel, tasks, chunksize=chunksize)
    reaction_id_equal_marg = sum(result) / float(reaction_pairs)

    # Marginal probability of observing two reactions with different ids.
    reaction_id_not_equal_marg = 1.0 - reaction_id_equal_marg

    reaction_id_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.reactions, model2.reactions,
        reaction_id_likelihood,
        reaction_prior, reaction_id_equal_marg, reaction_id_not_equal_marg)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()
    # Reaction name
    print('Calculating reaction name likelihoods...', end=' ')
    sys.stdout.flush()

    # Marginal probability of observing two reactions with the same name.
    tasks = ((util.name_equals, (r1.name, r2.name)) for r1, r2 in product(
        itervalues(model1.reactions),
        itervalues(model2.reactions)))
    result = pool.map(parallel_equel, tasks, chunksize=chunksize)
    reaction_name_equal_marg = sum(result) / float(reaction_pairs)

    reaction_name_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.reactions, model2.reactions,
        reaction_name_likelihood, reaction_prior, reaction_name_equal_marg)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()

    # Reaction equation

    print('Calculating reaction equation likelihoods...', end=' ')
    sys.stdout.flush()
    reaction_equation_likelihoods = pairwise_likelihood(
        pool, chunksize, model1.reactions, model2.reactions,
        reaction_equation_compound_mapping_likelihood,
        cpd_map, cpd_score, compartment_map)

    print('%.2f seconds' % (time.time() - t))
    t = time.time()

    # Reaction genes
    # For each gene, the marginal probability of observing that gene
    # in each model. We use this as an approximation of the probability of
    # observing a pair of genes in two reactions given that the reaction
    # do _not_ match.
    if gene:
        print('Calculating reaction genes likelihoods...', end=' ')
        sys.stdout.flush()

        # Marginal probability of observing two reactions with
        # equal gene associations.
        tasks = ((util.genes_equals, (r1.genes, r2.genes))
                 for r1, r2 in product(
                     itervalues(model1.reactions),
                     itervalues(model2.reactions)))
        result = pool.map(parallel_equel, tasks, chunksize=chunksize)
        reaction_genes_equal_marg = sum(result) / float(reaction_pairs)

        # Marginal probability of observing two reactions with unequal
        # gene associations.
        reaction_genes_not_equal_marg = 1.0 - reaction_genes_equal_marg - (
            sum(r1.genes is None or r2.genes is None
                for r1, r2 in product(itervalues(model1.reactions),
                                      itervalues(model2.reactions))) /
            reaction_pairs)

        reaction_genes_likelihoods = pairwise_likelihood(
            pool, chunksize, model1.reactions, model2.reactions,
            reaction_genes_likelihood,
            reaction_prior, reaction_genes_equal_marg,
            reaction_genes_not_equal_marg, gene_map)

        print('%.2f seconds' % (time.time() - t))
        t = time.time()
    else:
        reaction_genes_likelihoods = pairwise_likelihood(
            pool, chunksize, model1.reactions, model2.reactions,
            fake_likelihood)

    pool.close()
    pool.join()

    if log:
        merge_result = pd.merge(reaction_id_likelihoods,
                                reaction_name_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_id', '_name'))
        merge_result = pd.merge(merge_result, reaction_equation_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_name', '_equation'))
        merge_result = pd.merge(merge_result, reaction_genes_likelihoods,
                                left_index=True, right_index=True,
                                suffixes=('_equation', '_genes'))

        merge_result.to_csv(outpath + '/reaction_log.tsv', sep='\t')

    all_likelihoods = [reaction_id_likelihoods,
                       reaction_name_likelihoods,
                       reaction_equation_likelihoods,
                       reaction_genes_likelihoods]

    return (bayes_posterior(reaction_prior,
                            likelihood_products(all_likelihoods)),
            bayes_posterior(reaction_prior, reaction_id_likelihoods),
            bayes_posterior(reaction_prior, reaction_name_likelihoods),
            bayes_posterior(reaction_prior, reaction_equation_likelihoods),
            bayes_posterior(reaction_prior, reaction_genes_likelihoods))


def check_cpd_charge(compound, source):
    if compound.charge is not None:
        if isinstance(compound.charge, int):
            return True
        else:
            logger.warning(
                "Compound charge should be an integer, however, charge of "
                "compound '{}' in '{}' model is '{}', which is invalid. "
                "Please remove or fix it before running modelmapping "
                "command.".format(compound.id, source, compound.charge))
            return False



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2016  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016  Chao Liu <lcddzyx@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import operator
import logging

from six.moves import reduce

from .formula import Formula, ParseError

logger = logging.getLogger(__name__)


def reaction_charge(reaction, compound_charge):
    """Calculate the overall charge for the specified reaction.

    Args:
        reaction: :class:`psamm.reaction.Reaction`.
        compound_charge: a map from each compound to charge values.
    """

    charge_sum = 0.0
    for compound, value in reaction.compounds:
        charge = compound_charge.get(compound.name, float('nan'))
        charge_sum += charge * float(value)
    return charge_sum


def charge_balance(model):
    """Calculate the overall charge for all reactions in the model.

    Yield (reaction, charge) pairs.

    Args:
        model: :class:`psamm.datasource.native.NativeModel`.
    """

    compound_charge = {}
    for compound in model.compounds:
        if compound.charge is not None:
            compound_charge[compound.id] = compound.charge

    for reaction in model.reactions:
        charge = reaction_charge(reaction.equation, compound_charge)
        yield reaction, charge


def reaction_formula(reaction, compound_formula):
    """Calculate formula compositions for both sides of the specified reaction.

    If the compounds in the reaction all have formula, then calculate and
    return the chemical compositions for both sides, otherwise return `None`.

    Args:
        reaction: :class:`psamm.reaction.Reaction`.
        compound_formula: a map from compound id to formula.
    """

    def multiply_formula(compound_list):
        for compound, count in compound_list:
            yield count * compound_formula[compound.name]

    for compound, _ in reaction.compounds:
        if compound.name not in compound_formula:
            return None
    else:
        left_form = reduce(
            operator.or_, multiply_formula(reaction.left), Formula())
        right_form = reduce(
            operator.or_, multiply_formula(reaction.right), Formula())
    return left_form, right_form


def formula_balance(model):
    """Calculate formula compositions for each reaction.

    Call :func:`reaction_formula` for each reaction.
    Yield (reaction, result) pairs, where result has two formula compositions
    or `None`.

    Args:
        model: :class:`psamm.datasource.native.NativeModel`.
    """

    # Mapping from compound id to formula
    compound_formula = {}
    for compound in model.compounds:
        if compound.formula is not None:
            try:
                f = Formula.parse(compound.formula).flattened()
                compound_formula[compound.id] = f
            except ParseError as e:
                msg = 'Error parsing formula for compound {}:\n{}\n{}'.format(
                    compound.id, e, compound.formula)
                if e.indicator is not None:
                    msg += '\n{}'.format(e.indicator)
                logger.warning(msg)

    for reaction in model.reactions:
        yield reaction, reaction_formula(reaction.equation, compound_formula)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Jing Wang <wjingsjtu@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2021  Elysha Sameth <esameth1@uri.edu>

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import errno
import pandas as pd
import re
from future import standard_library
standard_library.install_aliases()


class Curator(object):
    """Parse and save mapping files during manual curation.

    Use :meth:`.add_mapping` to add new curated pairs. Save current progress
    into files by :meth:`.save`.

    Besides the curated mapping files, the :class:`Curator` will also store
    false mappings into `.false` files, and compounds and reactions
    to be ignored can be stored
    in `.ignore` files. For example, if the `curated_compound_map_file` is set
    to `curated_compound_mapping.tsv`, then the false mappings will be stored
    in `curated_compound_mapping.tsv.false`, and the pairs to be ignored
    should be stored in `curated_compound_mapping.tsv.ignore`.

    If the curated files already exist, the :class:`Curator` will consider them
    as the previous progress, then append new curation results.

    Args:
        compound_map_file: .tsv file of compound mapping result
        reaction_map_file: .tsv file of reaction mapping result
        curated_compound_map_file: .tsv file of curated compound mapping result
        curated_reaction_map_file: .tsv file of curated reaction mapping result
    """

    def __init__(self,
                 compound_map_file, reaction_map_file,
                 curated_compound_map_file, curated_reaction_map_file):
        self.compound_map_file = compound_map_file
        self.reaction_map_file = reaction_map_file
        self.curated_compound_map_file = curated_compound_map_file
        self.curated_reaction_map_file = curated_reaction_map_file
        self.false_compound_map_file = curated_compound_map_file + '.false'
        self.false_reaction_map_file = curated_reaction_map_file + '.false'
        self.ignore_compound_file = curated_compound_map_file + '.ignore'
        self.ignore_reaction_file = curated_reaction_map_file + '.ignore'

        self.compound_map = read_mapping(self.compound_map_file, [0, 1])
        self.compound_map.sort_values(by='p', inplace=True, ascending=False)
        self.reaction_map = read_mapping(self.reaction_map_file, [0, 1])
        self.reaction_map.sort_values(by='p', inplace=True, ascending=False)
        self.curated_compound_map = read_mapping(
            self.curated_compound_map_file, [0, 1])
        self.curated_reaction_map = read_mapping(
            self.curated_reaction_map_file, [0, 1])
        self.false_compound_map = read_mapping(
            self.false_compound_map_file, [0, 1])
        self.false_reaction_map = read_mapping(
            self.false_reaction_map_file, [0, 1])
        self.ignore_compound = read_ignore(self.ignore_compound_file)
        self.ignore_reaction = read_ignore(self.ignore_reaction_file)

        self.num_curated_compounds = len(self.curated_compound_map.index)
        self.num_curated_compounds_left = \
            self.compound_map.index.get_level_values(0).nunique() - \
            self.num_curated_compounds
        self.num_curated_reactions = len(self.curated_reaction_map.index)
        self.num_curated_reactions_left = \
            self.reaction_map.index.get_level_values(0).nunique() - \
            self.num_curated_reactions

    def reaction_checked(self, id):
        """Return True if reaction pair has been checked.

        Args:
            id: one reaction id or a tuple of id pair
        """
        return (id in self.curated_reaction_map.index or
                id in self.false_reaction_map.index)

    def reaction_ignored(self, id):
        """Return True if reaction id is in ignore list."""
        return id in self.ignore_reaction

    def compound_checked(self, id):
        """Return True if compound pair has been checked.

        Args:
            id: one compound id or a tuple of id pair
        """
        return (id in self.curated_compound_map.index or
                id in self.false_compound_map.index)

    def compound_ignored(self, id):
        """Return True if compound id is in ignore list."""
        return id in self.ignore_compound

    def add_mapping(self, id, type, correct):
        """Add new mapping result to curated list.

        Args:
            id: tuple of id pair
            type: 'c' for compound mapping, 'r' for reaction mapping
            correct: True if the mapping pair is correct
        """
        if type == 'c':
            if correct:
                self.curated_compound_map = add_mapping(
                    self.curated_compound_map,
                    self.compound_map.loc[id]
                )
            else:
                self.false_compound_map = add_mapping(
                    self.false_compound_map,
                    self.compound_map.loc[id]
                )
        if type == 'r':
            if correct:
                self.curated_reaction_map = add_mapping(
                    self.curated_reaction_map,
                    self.reaction_map.loc[id]
                )
            else:
                self.false_reaction_map = add_mapping(
                    self.false_reaction_map,
                    self.reaction_map.loc[id]
                )

    def add_ignore(self, id, type):
        """Add id to ignore list.

        Args:
            id: id to be ignored
            type: 'c' for compound mapping, 'r' for reaction mapping
        """
        if type == 'c':
            self.ignore_compound.append(id)
        if type == 'r':
            self.ignore_reaction.append(id)

    def save(self):
        """Save current curator to files."""
        if len(self.curated_compound_map) > 0:
            self.curated_compound_map.to_csv(
                self.curated_compound_map_file, sep='\t'
            )
        if len(self.curated_reaction_map) > 0:
            self.curated_reaction_map.to_csv(
                self.curated_reaction_map_file, sep='\t'
            )
        if len(self.false_compound_map) > 0:
            self.false_compound_map.to_csv(
                self.false_compound_map_file, sep='\t'
            )
        if len(self.false_reaction_map) > 0:
            self.false_reaction_map.to_csv(
                self.false_reaction_map_file, sep='\t'
            )
        if len(self.ignore_compound) > 0:
            write_ignore(self.ignore_compound, self.ignore_compound_file)
        if len(self.ignore_reaction) > 0:
            write_ignore(self.ignore_reaction, self.ignore_reaction_file)
        print('Progress saved\n')


def read_mapping(file, index_col):
    try:
        df = pd.read_csv(file, sep='\t', index_col=index_col)
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise
        df = pd.DataFrame()
    return df


def add_mapping(map, row):
    return map.append(pd.DataFrame(row).T)


def read_ignore(file):
    ignore = list()
    try:
        with open(file) as f:
            for r in f:
                ignore.append(r.strip())
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise
    return ignore


def write_ignore(ignore, file):
    with open(file, 'w') as o:
        for i in ignore:
            o.write('%s\n' % str(i))


def filter_search_term(s):
    return re.sub(r'[^a-z0-9]+', '', s.lower())


def search_compound(model, id):
    """Search a set of compounds, then print detailed properties.

    Args:
        id: a list of compound ids
    """
    selected_compounds = set()

    for compound in model.compounds:
        if len(id) > 0:
            if any(c == compound.id for c in id):
                selected_compounds.add(compound)
                continue

    # Show results
    for compound in selected_compounds:
        props = set(compound.properties) - {'id'}
        if compound.filemark is not None:
            print('Defined in {}'.format(compound.filemark))
        print('id: {}'.format(compound.id))
        for prop in sorted(props):
            print('{}: {}'.format(prop, compound.properties[prop]))
        print('\n')


def search_reaction(model, ids):
    """Search a set of reactions, print detailed properties, then return a
    generator. Each item in the generator is a list of compounds in the
    corresponding reaction.

    Args:
        ids: a list of reaction ids
    """
    selected_reactions = set()
    for r in ids:
        if r in model.reactions:
            selected_reactions.add(model.reactions[r])

    # Show results
    for reaction in selected_reactions:
        props = set(reaction.properties) - {'id', 'equation'}
        if reaction.filemark is not None:
            print('Defined in {}'.format(reaction.filemark))
        print('id: {}'.format(reaction.id))
        print('equation: {}'.format(
            reaction.equation))
        translated_equation = reaction.equation.translated_compounds(
            lambda c: model.compounds[c].name)
        if reaction.equation != translated_equation:
            print('equation (compound names): {}'.format(
                translated_equation))
        for prop in sorted(props):
            print('{}: {}'.format(prop, reaction.properties[prop]))
        print('\n')
        yield [c for c, v in reaction.equation.compounds]
